import abc
import contextlib
import time
import json
from collections import namedtuple

import paramiko
from libcloud.compute.base import NodeAuthSSHKey


NodeSize = namedtuple('NodeSize', 'id memory cpu disk monthly_cost extras')
Region = namedtuple('Region', 'id name')


class BaseLibcloudProvider(abc.ABC):
    username = 'root'
    driver = None

    def generate_ssh_key(self): # pylint: disable=no-self-use
        # paramiko doesn't yet support creating Ed25519 keys :'(
        return paramiko.ECDSAKey.generate()


    def create_remote_ssh_key(self, key_name, ssh_key, public_key):
        '''Return a tuple of (remote_key, auth_key)'''
        remote_key = self.driver.create_key_pair(key_name, public_key)
        auth_key = NodeAuthSSHKey(remote_key.public_key)
        return remote_key, auth_key


    @contextlib.contextmanager
    def create_temp_ssh_key(self, key_name):
        local_key = self.generate_ssh_key()
        # There's three different variants of the key here, the local key that
        # has the private part, the remote key which has the provider mapping to
        # delete it later, and the auth key, which is passed to the provider
        # again when creating the node to allow authentication.
        public_key = '%s %s' % (local_key.get_name(), local_key.get_base64())
        remote_key, auth_key = self.create_remote_ssh_key(key_name, local_key, public_key)
        print('Created temp ssh key')

        try:
            yield local_key, auth_key
        finally:
            print('Destroying %s ssh key' % self.__class__.__name__)
            self.destroy_remote_ssh_key(remote_key)


    def destroy_remote_ssh_key(self, remote_key):
        self.driver.delete_key_pair(remote_key)


    def wait_for_public_ip(self, node):
        if node.public_ips and node.public_ips[0] != '0.0.0.0':
            return node
        timeout = 180
        start_time = time.time()
        while True:
            time.sleep(2)
            node = self.get_updated_node(node)
            if node.public_ips and node.public_ips[0] != '0.0.0.0':
                return node
            if time.time() - start_time > timeout:
                raise ValueError('Timed out waiting for node IP: %s' % node.id)


    def get_updated_node(self, old_node):
        for node in self.driver.list_nodes():
            if node.id == old_node.id:
                return node

        raise ValueError('Updated node for %s not found' % old_node.id)


    def get_size(self, size_name):
        sizes = self.driver.list_sizes()
        for size in sizes:
            if size.name == size_name or size.id == size_name:
                # Allow targeting by id too
                return size

        raise ValueError('Unknown size: %s' % size_name)


    def get_location(self, location_id):
        for location in self.driver.list_locations():
            if location.id == location_id or location.name == location_id:
                return location

        raise ValueError('Location %s not found' % location_id)


    def destroy_node(self, node, extra=None):
        self.driver.destroy_node(node)


    def get_node(self, node_id):
        for node in self.driver.list_nodes():
            if node_id in (node.id, node.name):
                return node

        raise ValueError('No node with id %s found in provider %s' % (
            node_id, self.__class__.__name__))


    def wait_for_init_script(self, client, extra=None):
        # Creds to https://stackoverflow.com/a/14158100 for a way to get the pid
        _, stdout, stderr = client.exec_command('echo $$ && exec tail -f /var/log/cloud-init-output.log')
        tail_pid = int(stdout.readline())
        for line in stdout:
            print(line, end='')
            if line.startswith('Cloud-init') and ' finished ' in line:
                stdout.channel.close()
                client.exec_command('kill %d' % tail_pid)
                break

        for line in stderr:
            print('Cloud-init stderr: %s' % line.strip())

        _, stdout, stderr = client.exec_command('cat /run/cloud-init/result.json', timeout=3)
        if stdout.channel.recv_exit_status() != 0:
            raise ValueError('Failed to get cloud-init status')

        cloud_init_result = json.loads(''.join(stdout))
        if cloud_init_result['v1']['errors']:
            raise ValueError('cloud-init failed: %s' % ', '.join(cloud_init_result['v1']['errors']))


    def post_connect(self, hart_node):
        pass


    def get_sizes(self):
        # Very provider specific, implemented individually
        raise NotImplemented()


    def get_regions(self):
        regions = []
        for location in self.driver.list_locations():
            regions.append(Region(location.id, location.name))
        regions.sort(key=lambda r: r.name)
        return regions



import os

import toml

from .providers import provider_map


def build_provider_from_file(provider_alias, config_file='~/.hartrc', **kwargs):
    constructor = provider_map[provider_alias]
    file_path = os.path.expanduser(config_file)
    with open(file_path) as fh:
        config = toml.load(fh)

    provider_config = config['providers'][provider_alias]
    return constructor(**provider_config, **kwargs)



#!./venv/bin/python

import base64
import datetime
import json
import os
import subprocess
import sys
import time
import traceback
from collections import namedtuple

import yaml
from jinja2 import Template

from .ssh import get_verified_ssh_client, ssh_run_command
from .constants import DEBIAN_VERSIONS

HartNode = namedtuple('HartNode', 'minion_id public_ip node provider ssh_key ssh_canary node_extra')


def create_minion(
        minion_id,
        provider,
        region=None,
        size=None,
        salt_branch='latest',
        debian_codename='stretch',
        tags=None,
        private_networking=False,
        minion_config=None,
        zone=None,
        subnet=None,
        ):
    hart_node = create_node(
        minion_id,
        provider,
        region,
        size,
        salt_branch,
        debian_codename,
        tags,
        private_networking,
        minion_config,
        zone,
        subnet,
    )
    try:
        connect_minion(hart_node)
    except:
        sys.stderr.write('Destroying node since it failed to connect\n')
        hart_node.provider.destroy_node(hart_node.node, extra=hart_node.node_extra)
        raise


def connect_minion(hart_node):
    username = hart_node.provider.username
    with get_verified_ssh_client(
            hart_node.public_ip,
            hart_node.ssh_key,
            hart_node.ssh_canary,
            username) as client:
        hart_node.provider.wait_for_init_script(client, hart_node.node_extra)
        minion_pubkey = get_minion_pubkey(client, should_sudo=username != 'root')
        print(minion_pubkey)
        trust_minion_key(hart_node.minion_id, minion_pubkey)
        print('Minion added: %s' % hart_node.public_ip)
        verify_minion_connection(client, hart_node.minion_id, username)
        hart_node.provider.post_connect(hart_node)


def create_node(
        minion_id,
        provider,
        region=None,
        size=None,
        salt_branch='latest',
        debian_codename='stretch',
        tags=None,
        private_networking=False,
        minion_config=None,
        zone=None,
        subnet=None,
        ):
    ssh_canary = base64.b64encode(os.urandom(30)).decode('utf-8')
    cloud_init_template = get_cloud_init_template()
    master_pubkey = get_master_pubkey()
    default_minion_config = {
        'id': minion_id,
    }
    if minion_config is not None:
        default_minion_config.update(minion_config)

    cloud_init = cloud_init_template.render(**{
        'minion_config': yaml.dump(default_minion_config),
        'ssh_canary': ssh_canary,
        'salt_branch': salt_branch,
        'debian_version': DEBIAN_VERSIONS[debian_codename],
        'debian_codename': debian_codename,
        'master_pubkey': master_pubkey,
    })

    key_name = build_ssh_key_name(minion_id)

    if not check_existing_minion(minion_id):
        print('Existing minions were found and did want to overwrite, aborting')
        return

    with provider.create_temp_ssh_key(key_name) as (ssh_key, auth_key):
        node = None
        kwargs = {}
        if size:
            kwargs['size'] = size
        if zone:
            kwargs['zone'] = zone
        if subnet:
            kwargs['subnet'] = subnet
        try:
            node, extra = provider.create_node(
                minion_id,
                region,
                debian_codename,
                auth_key,
                cloud_init,
                private_networking,
                tags,
                **kwargs)
            node = provider.wait_for_public_ip(node)
            public_ip = node.public_ips[0]
            print('Node running at %s' % public_ip)
            return HartNode(minion_id, public_ip, node, provider, ssh_key, ssh_canary, extra)
        except:
            traceback.print_exc()
            if node:
                sys.stderr.write('Destroying node since it failed initialization\n')
                provider.destroy_node(node, extra)
            raise


def destroy_minion(minion_id, provider):
    disconnect_minion(minion_id)
    print('Destroying minion')
    node = provider.get_node(minion_id)
    provider.destroy_node(node)


def disconnect_minion(minion_id):
    print('Deleting the salt minion %s' % minion_id)
    subprocess.check_call([
        'salt-key',
        '--delete=%s' % minion_id,
        '--yes',
    ])


def destroy_node(hart_node):
    hart_node.provider.destroy_node(hart_node.node, extra=hart_node.node_extra)


def get_master_pubkey():
    with open('/etc/salt/pki/master/master.pub') as fh:
        return fh.read()


def check_existing_minion(minion_id):
    minions = json.loads(subprocess.check_output([
        'salt-key',
        '--print', minion_id,
        '--out=json',
    ]).decode('utf-8'))

    existing_categories = minions.keys()
    if existing_categories:
        should_continue = input('Existing minions matching %s were found in %s, overwrite? [y/N]' % (
            minion_id, ', '.join(existing_categories)))
        return should_continue == 'y'

    return True


def trust_minion_key(minion_id, minion_pubkey):
    with open('/etc/salt/pki/master/minions/%s' % minion_id, 'wb') as fh:
        fh.write(minion_pubkey.encode('utf-8'))

    # If the minion connected before we trusted the key, remove the duplicate key in minions_pre
    pre_key_path = '/etc/salt/pki/master/minions_pre/%s' % minion_id
    try:
        pre_fh = open(pre_key_path)
    except IOError:
        return

    with pre_fh:
        pre_key = pre_fh.read()

    if pre_key.strip() == minion_pubkey.strip():
        print('Got early connection attempt from minion, removing pre-accept key')
        os.remove(pre_key_path)


def verify_minion_connection(client, minion_id, username):
    # The restart is needed since the minion might have attempted connecting to
    # the salt master before the key got trusted and thus might not be ready to
    # respond to a ping from the master
    prefix = 'sudo ' if username != 'root' else ''
    ssh_run_command(client,
        '{0}salt-call test.ping && {0}service salt-minion restart'.format(prefix),
        timeout=15)

    # Give the minion some time to start before attempting another ping
    time.sleep(5)

    # Also test that the master can reach the minion
    subprocess.check_call(['salt', minion_id, 'test.ping'])

    authorized_keys_path = '/root/.ssh/authorized_keys'
    if username != 'root':
        authorized_keys_path = '/home/%s/.ssh/authorized_keys' % username
    ssh_run_command(client, 'rm %s' % authorized_keys_path)


def get_cloud_init_template(template_name='minion.sh'):
    template_path = os.path.join(os.path.dirname(__file__), 'cloud-init', template_name)
    with open(template_path) as fh:
        cloud_init_template = Template(fh.read())
    return cloud_init_template


def build_ssh_key_name(minion_id):
    current_date = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H-%M-%S')
    return 'temp-for-%s-at-%s' % (minion_id, current_date)


def get_minion_pubkey(client, should_sudo):
    _, stdout, stderr = client.exec_command('%scat /etc/salt/pki/minion/minion.pub' % ('sudo ' if should_sudo else '',), timeout=3)
    if stdout.channel.recv_exit_status() != 0:
        raise ValueError('Failed to get minion pubkey: %s' % ''.join(stderr))

    return ''.join(stdout)



from .minions import (
    create_minion,
    destroy_minion,
    create_node,
    destroy_node,
    connect_minion,
    disconnect_minion,
)
from .config import build_provider_from_file
from .constants import DEBIAN_VERSIONS



import base64
import contextlib
import hashlib
import time
import sys

import paramiko


class IgnorePolicy(paramiko.MissingHostKeyPolicy):
    def missing_host_key(self, client, hostname, key):
        # key.__str__() isn't valid according to str() since it returns bytes, thus calling it directly
        fingerprint = 'SHA256:%s' % base64.b64encode(hashlib.sha256(key.__str__()).digest()).decode('utf-8')
        print('Accepting host key type %s with fingerprint %s' % (key.get_name(), fingerprint))


def ssh_run_command(client, command, timeout=3, sensitive=False):
    captured_stdout = []
    captured_stderr = []
    _, stdout, stderr = client.exec_command(command, timeout=timeout)
    # while not stdout.channel.exit_status_ready():
    # This causes stderr to be depleted before doing anything else, for true interactive output
    # we need to use the channel API directly and use recv_stderr_ready/recv_
    for chunk in stderr:
        captured_stderr.append(chunk)
        end = '' if chunk[-1] == '\n' else '\n'
        sys.stderr.write('stderr: %s%s' % (chunk, end))
    for chunk in stdout:
        captured_stdout.append(chunk)
        end = '' if chunk[-1] == '\n' else '\n'
        print('stdout: %s' % chunk, end=end)

    if stdout.channel.recv_exit_status() != 0:
        if sensitive:
            raise ValueError('Sensitive command failed with exit code %d' % stdout.channel.recv_exit_status())
        raise ValueError('Command %s failed with exit code %d' % (command, stdout.channel.recv_exit_status()))

    return ''.join(captured_stdout)


@contextlib.contextmanager
def get_verified_ssh_client(ip, ssh_key, canary, username='root'):
    client = connect_to_droplet(ip, ssh_key, username)
    print('Connected')

    # Verify the ssh canary as the first thing to not run any potentially
    # dangerous operations on an untrusted box
    wait_for_verified_ssh_canary(client, canary, should_sudo=username != 'root')
    print('Verified connection')
    try:
        yield client
    except:
        client.close()
        raise


def log_action(action, start_time):
    print('action=%s time=%.2fs' % (action, time.time() - start_time))


def connect_to_droplet(ip, client_ssh_key, username):
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(IgnorePolicy())
    timeout = 90
    start_time = time.time()
    while time.time() - start_time < timeout:
        try:
            client.connect(ip, username=username, pkey=client_ssh_key, timeout=3)
            log_action('connect', start_time)
            break
        except Exception as e:
            print('Could not connect yet, waiting (%s)' % e)
            time.sleep(2)
    else:
        raise ValueError('Failed to connect to new node')

    return client


def wait_for_verified_ssh_canary(client, ssh_canary, should_sudo):
    timeout = 15
    start_time = time.time()
    while time.time() - start_time < timeout:
        # The remove is just a matter of cleanup, the canary isn't sensitive
        _, stdout, stderr = client.exec_command('cat /tmp/ssh-canary && {0}rm /tmp/ssh-canary'.format('sudo ' if should_sudo else ''), timeout=3)
        if stderr.channel.recv_exit_status() != 0:
            print('No ssh canary yet, waiting (%s)' % ''.join(stderr).strip())
            time.sleep(1)
            continue

        found_canary = ''.join(stdout).strip()
        if found_canary != ssh_canary:
            raise ValueError('ssh canary check failed!')

        log_action('canary-check', start_time)
        break
    else:
        raise ValueError('No canary found before timeout')



import ipaddress
import json
import datetime
import sys

import boto3
import ifaddr
import paramiko
from libcloud.compute.base import Node
from libcloud.compute.providers import get_driver
from libcloud.compute.types import Provider

from .base import BaseLibcloudProvider, NodeSize, Region


# The pricing API is a supreme clusterfuck that requires lots of special care.
# One particular issue we run into when listing EC2 instance sizes in a given
# region is that the API doesn't allow you to filter on region, but on
# "location", which is the human-readable name of the region. There's no API to
# get the mapping between regions and the name, thus you either have to hardcode
# it like we do here and be forced to push and update whenever Amazon adds
# another region, or scrape the
# https://docs.aws.amazon.com/general/latest/gr/rande.html#ec2_region page,
# which will break if they change the html structure of that page. Opted for the
# hardcoding here for simplicity.
region_to_location_map = {
    'us-east-2': 'US East (Ohio)',
    'us-east-1': 'US East (N. Virginia)',
    'us-west-1': 'US West (N. California)',
    'us-west-2': 'US West (Oregon)',
    'ap-east-1': 'Asia Pacific (Hong Kong)',
    'ap-south-1': 'Asia Pacific (Mumbai)',
    'ap-northeast-3': 'Asia Pacific (Osaka-Local)',
    'ap-northeast-2': 'Asia Pacific (Seoul)',
    'ap-southeast-1': 'Asia Pacific (Singapore)',
    'ap-southeast-2': 'Asia Pacific (Sydney)',
    'ap-northeast-1': 'Asia Pacific (Tokyo)',
    'ca-central-1': 'Canada (Central)',
    'cn-north-1': 'China (Beijing)',
    'cn-northwest-1': 'China (Ningxia)',
    'eu-central-1': 'EU (Frankfurt)',
    'eu-west-1': 'EU (Ireland)',
    'eu-west-2': 'EU (London)',
    'eu-west-3': 'EU (Paris)',
    'eu-north-1': 'EU (Stockholm)',
    'sa-east-1': 'South America (Sao Paulo)',
    'me-south-1': 'Middle East (Bahrain)',
}


class EC2Provider(BaseLibcloudProvider):
    username = 'admin'

    def __init__(self, aws_access_key_id, aws_secret_access_key, region):
        self.ec2 = boto3.client('ec2',
            region_name=region,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.region = region


    def generate_ssh_key(self):
        # EC2 only supports RSA for ssh keys
        return paramiko.RSAKey.generate(2048)


    def create_remote_ssh_key(self, key_name, ssh_key, public_key):
        import_response = self.ec2.import_key_pair(KeyName=key_name, PublicKeyMaterial=public_key)
        assert import_response['ResponseMetadata']['HTTPStatusCode'] == 200
        return key_name, key_name


    def destroy_remote_ssh_key(self, remote_key):
        self.ec2.delete_key_pair(KeyName=remote_key)


    def create_node(self,
            minion_id,
            region,
            debian_codename,
            auth_key,
            cloud_init,
            private_networking,
            tags,
            size='t3.micro',
            zone=None,
            subnet=None):
        if not zone:
            raise ValueError('You must specify the ec2 availability zone')

        size = self.get_size(size)
        image = self.get_image(debian_codename)

        subnet_ids = [subnet] if subnet else []
        subnet_response = self.ec2.describe_subnets(SubnetIds=subnet_ids,
            Filters=[{'Name': 'availability-zone', 'Values': [zone]}])
        subnets = subnet_response['Subnets']

        if not subnets and subnet:
            raise ValueError('No subnet matching %s in %s' % (subnet, zone))
        elif not subnets:
            raise ValueError('No available subnets in %s' % zone)
        elif len(subnets) > 1:
            raise ValueError('More than one subnet in availability zone, specify'
                ' which one to use: %s' % (', '.join(s.id for s in subnets)))

        subnet = subnets[0]
        temp_security_group = self.create_temp_security_group(minion_id)
        create_response = self.ec2.run_instances(
            ImageId=image['ImageId'],
            InstanceType=size,
            KeyName=auth_key,
            Placement={'AvailabilityZone': zone},
            UserData=cloud_init,
            MinCount=1,
            MaxCount=1,
            NetworkInterfaces=[{
                'AssociatePublicIpAddress': True,
                'DeleteOnTermination': True,
                'DeviceIndex': 0,
                'SubnetId': subnet['SubnetId'],
                'Groups': [temp_security_group],
            }],
            TagSpecifications=[{
                'ResourceType': 'instance',
                'Tags': [{'Key': 'Name', 'Value': minion_id}],
            }],
        )
        instance = create_response['Instances'][0]

        node = Node(id=instance['InstanceId'], name=minion_id, state=instance['State']['Name'],
            public_ips=[], private_ips=[instance['PrivateIpAddress']],
            driver=self.ec2, created_at=instance['LaunchTime'], extra=None)
        return node, temp_security_group


    def get_updated_node(self, old_node):
        instance_response = self.ec2.describe_instances(InstanceIds=[old_node.id])
        instance = instance_response['Reservations'][0]['Instances'][0]
        public_ip = instance.get('PublicIpAddress')
        public_ips = [public_ip] if public_ip else []
        name = None
        for tag in instance['Tags']:
            if tag['Key'] == 'Name':
                name = tag['Value']
                break
        return Node(id=instance['InstanceId'], name=name, state=instance['State']['Name'],
            public_ips=public_ips, private_ips=[instance['PrivateIpAddress']],
            driver=self.ec2, created_at=instance['LaunchTime'], extra=None)


    def create_temp_security_group(self, minion_id):
        current_date = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H-%M-%S')
        name = 'temp-for-%s-%s' % (minion_id, current_date)

        # Get the external IPs for the current host
        external_ips = list(get_host_public_ips())
        if not external_ips:
            raise ValueError('Could not find any public IPs on the current '
                'host and thus wont be able to connect to the new node')

        group = self.ec2.create_security_group(
            GroupName=name,
            Description='Temporary group for initial saltmaster ssh initialization',
        )
        group_id = group['GroupId']

        self.ec2.authorize_security_group_ingress(
            GroupId=group_id,
            IpPermissions=[{
                'IpProtocol': 'tcp',
                'FromPort': 22,
                'ToPort': 22,
                'IpRanges': [{'CidrIp': ip} for ip in external_ips],
            }]
        )

        return group_id


    def get_image(self, debian_codename):
        official_debian_account = '379101102735'
        image_response = self.ec2.describe_images(Owners=[official_debian_account], Filters=[{
            'Name': 'architecture',
            'Values': ['x86_64'],
        }])
        dist_images = [image for image in image_response['Images'] if image['Name'].startswith('debian-%s-' % debian_codename)]
        dist_images.sort(key=lambda i: i['Name'])
        return dist_images[-1]


    def post_connect(self, hart_node):
        # Delete the temp security group that allowed ssh
        # Detach the security group from the instance. An instance must have at
        # least one security group, so we attach the VPC default group.
        self.delete_node_security_group(hart_node.node, hart_node.node_extra)


    def delete_node_security_group(self, node, group_id):
        print('Deleting temporary security group')
        response = self.ec2.describe_security_groups(GroupNames=['default'])
        default_group = response['SecurityGroups'][0]
        self.ec2.modify_instance_attribute(InstanceId=node.id, Groups=[default_group['GroupId']])
        response = self.ec2.delete_security_group(GroupId=group_id)


    def destroy_node(self, node, extra=None):
        if extra is not None:
            self.delete_node_security_group(node, extra)

        self.ec2.terminate_instances(InstanceIds=[node.id])


    def get_size(self, size_name):
        return size_name


    def get_sizes(self):
        pricing = boto3.client('pricing',
            region_name='us-east-1',
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
        )
        sizes = []
        location = region_to_location_map[self.region]
        filters = [
            {'Field': 'currentGeneration', 'Value': 'Yes', 'Type': 'TERM_MATCH'},
            {'Field': 'operatingSystem', 'Value': 'Linux', 'Type': 'TERM_MATCH'},
            {'Field': 'location', 'Value': location, 'Type': 'TERM_MATCH'},
            {'Field': 'productFamily', 'Value': 'Compute Instance', 'Type': 'TERM_MATCH'},
            {'Field': 'capacityStatus', 'Value': 'Used', 'Type': 'TERM_MATCH'},
            {'Field': 'preInstalledSw', 'Value': 'NA', 'Type': 'TERM_MATCH'},
            {'Field': 'tenancy', 'Value': 'Shared', 'Type': 'TERM_MATCH'},
        ]
        paginator = pricing.get_paginator('get_products')
        results = paginator.paginate(
                ServiceCode='AmazonEC2',
                Filters=filters,
                FormatVersion='aws_v1',
        )
        for response in results:
            for stringified_data in response['PriceList']:
                data = json.loads(stringified_data)
                attributes = data['product']['attributes']
                extras = {
                    'family': attributes['instanceFamily'],
                    'network': attributes['networkPerformance'],
                }
                if 'cpuFreq' in attributes:
                    extras['cpuFreq'] = attributes['clockSpeed']

                offer = data['terms']['OnDemand'].popitem()[1]
                price_dimension = offer['priceDimensions'].popitem()[1]
                price_per_unit = float(price_dimension['pricePerUnit']['USD'])
                price_unit = price_dimension['unit']
                if price_unit == 'Hrs':
                    price = price_per_unit * 720
                else:
                    raise ValueError('Unknown price unit: %s' % price_unit)

                memory_value, memory_unit = attributes['memory'].split()
                if memory_unit == 'GiB':
                    memory = float(memory_value.replace(',', ''))
                else:
                    raise ValueError('unknown memory unit: %s' % memory_unit)

                sizes.append(NodeSize(
                    attributes['instanceType'],
                    memory,
                    int(attributes['vcpu']),
                    attributes['storage'],
                    price,
                    extras,
                ))

        sizes.sort(key=lambda s: s.monthly_cost)
        return sizes


    def get_regions(self):
        regions = []
        response = self.ec2.describe_regions()
        for region in response['Regions']:
            region_boto = boto3.client('ec2',
                region_name=region['RegionName'],
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
            )
            az_response = region_boto.describe_availability_zones()
            for zone in az_response['AvailabilityZones']:
                # Don't fail if we don't know the name of the region to avoid
                # crashing for new regions we don't know about yet
                region_name = region_to_location_map.get(zone['RegionName'], 'Unknown')
                regions.append(Region(zone['ZoneName'], region_name))
        regions.sort(key=lambda r: r.name)
        return regions


def get_host_public_ips():
    for adapter in ifaddr.get_adapters():
        for ip in adapter.ips:
            string_ip = ip.ip[0] if isinstance(ip.ip, tuple) else ip.ip
            address = ipaddress.ip_address(string_ip)
            if address.is_global:
                yield '%s/%s' % (string_ip, address.max_prefixlen)



import base64
import hashlib

from libcloud.compute.providers import get_driver
from libcloud.compute.types import Provider

from ..constants import DEBIAN_VERSIONS
from .base import BaseLibcloudProvider, NodeSize


class DOProvider(BaseLibcloudProvider):

    def __init__(self, token, **kwargs):
        constructor = get_driver(Provider.DIGITAL_OCEAN)
        self.driver = constructor(token, api_version='v2')


    def create_node(self,
            minion_id,
            region,
            debian_codename,
            auth_key,
            cloud_init,
            private_networking,
            tags,
            size='s-1vcpu-1gb',
            **kwargs):
        key_fingerprint = pubkey_to_fingerprint(auth_key.pubkey)
        size = self.get_size(size)
        image = self.get_image(debian_codename)
        location = self.get_location(region)
        node = self.driver.create_node(minion_id, size, image, location, ex_user_data=cloud_init, ex_create_attr={
            'ssh_keys': [key_fingerprint],
            'private_networking': private_networking,
            'tags': tags,
        })
        return node, None


    def get_image(self, debian_codename):
        target_image = 'debian-%d-x64' % DEBIAN_VERSIONS[debian_codename]
        for image in self.driver.list_images():
            if image.extra['slug'] == target_image:
                return image
        raise ValueError('Image for %s not found' % debian_codename)


    def get_sizes(self):
        sizes = []
        for size in self.driver.list_sizes():
            sizes.append(NodeSize(
                size.name,
                size.ram/2**10,
                size.extra['vcpus'],
                '%d GB SSD' % size.disk,
                size.price*720,
                {'regions': size.extra['regions']},
            ))

        sizes.sort(key=lambda s: s.monthly_cost)
        return sizes


def pubkey_to_fingerprint(pubkey):
    '''Encodes arbitrary binary data as colon-separated hex pairs, like ba:5e:ba:11.'''
    base64_bytes = pubkey.split(' ')[1]
    key_blob = base64.b64decode(base64_bytes)
    hex_encoded = hashlib.md5(key_blob).hexdigest()
    return ':'.join(a + b for a, b in zip(hex_encoded[::2], hex_encoded[1::2]))



import contextlib
import datetime
import time

from libcloud.compute.providers import get_driver
from libcloud.compute.types import Provider, NodeState
from libcloud.utils.py3 import httplib

from .base import BaseLibcloudProvider, NodeSize


class VultrProvider(BaseLibcloudProvider):

    def __init__(self, token, **kwargs):
        constructor = get_driver(Provider.VULTR)
        self.driver = constructor(token)


    def create_node(self,
            minion_id,
            region,
            debian_codename,
            auth_key,
            cloud_init,
            private_networking,
            tags,
            size='201'):
        size = self.get_size(size)
        image = self.get_image(debian_codename)
        location = self.get_location(region)
        # Vultr has replaced cloud-init with their own startup script
        # implementation. This doesn't seem to be reflected in their docs,
        # but a comment here indicates as much:
        # https://discuss.vultr.com/discussion/582/cloud-init-user-data-testing/p3
        script_id = self.create_temp_startup_script(minion_id, cloud_init)
        node_extra = {
            'script_id': script_id,
        }
        node = self.driver.create_node(minion_id, size, image, location, ex_ssh_key_ids=[
            auth_key.id
        ], ex_create_attr={
            'script_id': script_id,
            'notify_activate': False,
            'enable_private_network': 'yes' if private_networking else 'no',
            'hostname': minion_id,
            'tag': tags,
        })
        # Vultr has a race condition where if the ssh key is deleted too early,
        # ie before the node has read it on startup, it won't be available to
        # use for logging in. Thus we delay the return here until the node state
        # indicates it's ready to continue.
        start_time = time.time()
        while time.time() - start_time < 180:
            print('Waiting for node to boot (%s)' % node.state)
            time.sleep(2)
            node = self.get_updated_node(node)
            if node.state != NodeState.PENDING:
                print('Node in state %s, continuing' % node.state)
                break
        else:
            # Can't auto-destroy since the API doesn't enable destroying nodes
            # before they are initialized
            raise ValueError('Failed to start node before timeout')

        return node, node_extra


    def create_temp_startup_script(self, minion_id, cloud_init):
        current_date = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H-%M-%S')
        params = {
            'name': 'temp-script-for-%s-%s' % (minion_id, current_date),
            'script': cloud_init,
        }

        result = self.driver.connection.post('/v1/startupscript/create', params)
        if result.status != httplib.OK:
            raise ValueError('Failed to create temp startupscript')

        script_id = result.object['SCRIPTID']
        return script_id


    def delete_startup_script(self, node_extra):
        if not node_extra['script_id']:
            # The delete might be called multiple times on error, prevent it
            # from trying to delete the script twice
            return

        script_id = node_extra['script_id']
        node_extra['script_id'] = None

        print('Start up script with id %d' % script_id)

        params = {'SCRIPTID': script_id}
        result = self.driver.connection.post('/v1/startupscript/destroy', params)
        if result.status != httplib.OK:
            print('Failed to delete temp startupscript %s' % script_id)


    def create_remote_ssh_key(self, key_name, ssh_key, public_key):
        # The vultr provider mistakenly returns a success bool instead of the key pair.
        # PR from 2017 that fixes it: https://github.com/apache/libcloud/pull/998
        # ¯\_(ツ)_/¯s
        self.driver.create_key_pair(key_name, public_key)
        key_pairs = self.driver.list_key_pairs()
        for key_pair in key_pairs:
            if key_pair.name == key_name:
                break
        else:
            raise ValueError('Failed to create ssh key pair')
        return key_pair, key_pair


    def get_image(self, debian_codename):
        for image in self.driver.list_images():
            if image.extra['family'] == 'debian' and image.extra['arch'] == 'x64' and debian_codename in image.name:
                return image

        raise ValueError('Debian %s image not found' % debian_codename)


    def destroy_node(self, node, extra=None):
        if extra is not None:
            self.delete_startup_script(extra)

        # Vultr doesn't handle deleting nodes that haven't finished
        # initialization well, wait for the node to finish boot before
        # destroying it
        timeout = 180
        start_time = time.time()
        while True:
            time.sleep(3)
            node = self.get_updated_node(node)
            if node.state == NodeState.RUNNING:
                self.driver.destroy_node(node)
                break
            if time.time() - start_time > timeout:
                raise ValueError('Timed out waiting to delete initializing node: %s' % node.id)


    def wait_for_init_script(self, client, node_extra):
        # If we delete the startup script any earlier it might not get onto the node and
        # boot might fail
        self.delete_startup_script(node_extra)

        # Creds to https://stackoverflow.com/a/14158100 for a way to get the pid
        _, stdout, stderr = client.exec_command('echo $$ && exec tail -f /tmp/firstboot.log')
        tail_pid = int(stdout.readline())
        for line in stdout:
            print(line, end='')
            if line.startswith('Hart init script complete'):
                stdout.channel.close()
                client.exec_command('kill %d' % tail_pid)
                return

        raise ValueError('Failed to complete init script: %s' % ''.join(stderr).strip())


    def get_sizes(self):
        sizes = []
        for size in self.driver.list_sizes():
            sizes.append(NodeSize(
                size.id,
                size.ram/2**10,
                size.extra['vcpu_count'],
                '%d GB SSD' % size.disk,
                size.price,
                {},
            ))

        sizes.sort(key=lambda s: s.monthly_cost)
        return sizes
